@@ -2,7 +2,7 @@
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
-use std::collections::HashMap;
+use std::collections::{BTreeMap, HashMap};
 use std::default::Default;
 use std::str::FromStr;
 use std::{fmt, fs, path::Path as StdPath};
@@ -10,8 +10,6 @@ use std::{fmt, fs, path::Path as StdPath};
 use serde::de::value::{MapAccessDeserializer, SeqAccessDeserializer};
 use serde::de::{Deserialize, Deserializer, MapAccess, SeqAccess, Visitor};
 
-use toml;
-
 use crate::bindgen::ir::annotation::AnnotationSet;
 use crate::bindgen::ir::path::Path;
 use crate::bindgen::ir::repr::ReprAlign;
@@ -24,6 +22,7 @@ pub const VERSION: &str = env!("CARGO_PKG_VERSION");
 pub enum Language {
     Cxx,
     C,
+    Cython,
 }
 
 impl FromStr for Language {
@@ -41,6 +40,8 @@ impl FromStr for Language {
             "C++" => Ok(Language::Cxx),
             "c" => Ok(Language::C),
             "C" => Ok(Language::C),
+            "cython" => Ok(Language::Cython),
+            "Cython" => Ok(Language::Cython),
             _ => Err(format!("Unrecognized Language: '{}'.", s)),
         }
     }
@@ -48,6 +49,71 @@ impl FromStr for Language {
 
 deserialize_enum_str!(Language);
 
+impl Language {
+    pub(crate) fn typedef(self) -> &'static str {
+        match self {
+            Language::Cxx | Language::C => "typedef",
+            Language::Cython => "ctypedef",
+        }
+    }
+}
+
+/// Controls what type of line endings are used in the generated code.
+#[derive(Debug, Clone, Copy)]
+#[allow(clippy::upper_case_acronyms)]
+pub enum LineEndingStyle {
+    /// Use Unix-style linefeed characters
+    LF,
+    /// Use classic Mac-style carriage-return characters
+    CR,
+    /// Use Windows-style carriage-return and linefeed characters
+    CRLF,
+    /// Use the native mode for the platform: CRLF on Windows, LF everywhere else.
+    Native,
+}
+
+impl Default for LineEndingStyle {
+    fn default() -> Self {
+        LineEndingStyle::LF
+    }
+}
+
+impl LineEndingStyle {
+    pub fn as_str(&self) -> &'static str {
+        match self {
+            Self::LF => "\n",
+            Self::CR => "\r",
+            Self::CRLF => "\r\n",
+            Self::Native => {
+                #[cfg(target_os = "windows")]
+                {
+                    Self::CRLF.as_str()
+                }
+                #[cfg(not(target_os = "windows"))]
+                {
+                    Self::LF.as_str()
+                }
+            }
+        }
+    }
+}
+
+impl FromStr for LineEndingStyle {
+    type Err = String;
+
+    fn from_str(s: &str) -> Result<Self, Self::Err> {
+        match s.to_lowercase().as_ref() {
+            "native" => Ok(Self::Native),
+            "lf" => Ok(Self::LF),
+            "crlf" => Ok(Self::CRLF),
+            "cr" => Ok(Self::CR),
+            _ => Err(format!("Unrecognized line ending style: '{}'.", s)),
+        }
+    }
+}
+
+deserialize_enum_str!(LineEndingStyle);
+
 /// A style of braces to use for generating code.
 #[derive(Debug, Clone, PartialEq)]
 pub enum Braces {
@@ -125,6 +191,27 @@ impl FromStr for DocumentationStyle {
 
 deserialize_enum_str!(DocumentationStyle);
 
+/// How much of the documentation to include in the header file.
+#[derive(Debug, Clone, Copy)]
+pub enum DocumentationLength {
+    Short,
+    Full,
+}
+
+impl FromStr for DocumentationLength {
+    type Err = String;
+
+    fn from_str(s: &str) -> Result<DocumentationLength, Self::Err> {
+        match s.to_lowercase().as_ref() {
+            "short" => Ok(DocumentationLength::Short),
+            "full" => Ok(DocumentationLength::Full),
+            _ => Err(format!("Unrecognized documentation style: '{}'.", s)),
+        }
+    }
+}
+
+deserialize_enum_str!(DocumentationLength);
+
 /// A style of Style to use when generating structs and enums.
 #[derive(Debug, Copy, Clone, PartialEq)]
 pub enum Style {
@@ -147,6 +234,21 @@ impl Style {
             Style::Tag => false,
         }
     }
+
+    // https://cython.readthedocs.io/en/latest/src/userguide/external_C_code.html#styles-of-struct-union-and-enum-declaration
+    pub fn cython_def(self) -> &'static str {
+        if self.generate_tag() {
+            "cdef "
+        } else {
+            "ctypedef "
+        }
+    }
+}
+
+impl Default for Style {
+    fn default() -> Self {
+        Style::Both
+    }
 }
 
 impl FromStr for Style {
@@ -202,7 +304,7 @@ impl FromStr for ItemType {
 deserialize_enum_str!(ItemType);
 
 /// Type which specifies the sort order of functions
-#[derive(Debug, Clone, PartialEq)]
+#[derive(Debug, Clone, Copy, PartialEq)]
 pub enum SortKey {
     Name,
     None,
@@ -246,6 +348,20 @@ pub struct ExportConfig {
     pub item_types: Vec<ItemType>,
     /// Whether renaming overrides or extends prefixing.
     pub renaming_overrides_prefixing: bool,
+    /// Mangling configuration.
+    pub mangle: MangleConfig,
+}
+
+/// Mangling-specific configuration.
+#[derive(Debug, Clone, Deserialize, Default)]
+#[serde(rename_all = "snake_case")]
+#[serde(deny_unknown_fields)]
+#[serde(default)]
+pub struct MangleConfig {
+    /// The rename rule to apply to the type names mangled.
+    pub rename_types: RenameRule,
+    /// Remove the underscores used for name mangling.
+    pub remove_underscores: bool,
 }
 
 impl ExportConfig {
@@ -269,7 +385,7 @@ impl ExportConfig {
             }
         }
         if let Some(ref prefix) = self.prefix {
-            item_name.insert_str(0, &prefix);
+            item_name.insert_str(0, prefix);
         }
     }
 }
@@ -312,11 +428,13 @@ pub struct FunctionConfig {
     /// The style to layout the args
     pub args: Layout,
     /// The rename rule to apply to function args
-    pub rename_args: Option<RenameRule>,
+    pub rename_args: RenameRule,
     /// An optional macro to use when generating Swift function name attributes
     pub swift_name_macro: Option<String>,
-    /// Sort key for function names
-    pub sort_by: SortKey,
+    /// Sort key for functions
+    pub sort_by: Option<SortKey>,
+    /// Optional text to output after functions which return `!`.
+    pub no_return: Option<String>,
 }
 
 impl Default for FunctionConfig {
@@ -326,9 +444,10 @@ impl Default for FunctionConfig {
             postfix: None,
             must_use: None,
             args: Layout::Auto,
-            rename_args: None,
+            rename_args: RenameRule::None,
             swift_name_macro: None,
-            sort_by: SortKey::Name,
+            sort_by: None,
+            no_return: None,
         }
     }
 }
@@ -356,7 +475,7 @@ impl FunctionConfig {
 #[serde(default)]
 pub struct StructConfig {
     /// The rename rule to apply to the name of struct fields
-    pub rename_fields: Option<RenameRule>,
+    pub rename_fields: RenameRule,
     /// Whether to generate a constructor for the struct (which takes
     /// arguments to initialize all the members)
     pub derive_constructor: bool,
@@ -372,6 +491,8 @@ pub struct StructConfig {
     pub derive_gt: bool,
     /// Whether to generate a greater than or equal to operator on structs with one field
     pub derive_gte: bool,
+    /// Whether to generate a ostream serializer for the struct
+    pub derive_ostream: bool,
     /// Whether associated constants should be in the body. Only applicable to
     /// non-transparent structs, and in C++-only.
     pub associated_constants_in_body: bool,
@@ -422,6 +543,12 @@ impl StructConfig {
         }
         self.derive_gte
     }
+    pub(crate) fn derive_ostream(&self, annotations: &AnnotationSet) -> bool {
+        if let Some(x) = annotations.bool("derive-ostream") {
+            return x;
+        }
+        self.derive_ostream
+    }
 }
 
 /// Settings to apply to generated enums.
@@ -431,7 +558,7 @@ impl StructConfig {
 #[serde(default)]
 pub struct EnumConfig {
     /// The rename rule to apply to the name of enum variants
-    pub rename_variants: Option<RenameRule>,
+    pub rename_variants: RenameRule,
     /// Whether to add a `Sentinel` value at the end of every enum
     /// This is useful in Gecko for IPC serialization
     pub add_sentinel: bool,
@@ -459,6 +586,8 @@ pub struct EnumConfig {
     /// This is only generated if a copy constructor for the same tagged enum is
     /// generated as well.
     pub derive_tagged_enum_copy_assignment: bool,
+    /// Whether to generate a ostream serializer for the struct
+    pub derive_ostream: bool,
     /// Declare the enum as an enum class.
     /// Only relevant when targeting C++.
     pub enum_class: bool,
@@ -470,7 +599,7 @@ pub struct EnumConfig {
 impl Default for EnumConfig {
     fn default() -> EnumConfig {
         EnumConfig {
-            rename_variants: None,
+            rename_variants: RenameRule::None,
             add_sentinel: false,
             prefix_with_name: false,
             derive_helper_methods: false,
@@ -481,6 +610,7 @@ impl Default for EnumConfig {
             derive_tagged_enum_destructor: false,
             derive_tagged_enum_copy_constructor: false,
             derive_tagged_enum_copy_assignment: false,
+            derive_ostream: false,
             enum_class: true,
             private_default_tagged_enum_constructor: false,
         }
@@ -530,6 +660,12 @@ impl EnumConfig {
         }
         self.derive_tagged_enum_copy_assignment
     }
+    pub(crate) fn derive_ostream(&self, annotations: &AnnotationSet) -> bool {
+        if let Some(x) = annotations.bool("derive-ostream") {
+            return x;
+        }
+        self.derive_ostream
+    }
     pub(crate) fn enum_class(&self, annotations: &AnnotationSet) -> bool {
         if let Some(x) = annotations.bool("enum-class") {
             return x;
@@ -557,6 +693,8 @@ pub struct ConstantConfig {
     pub allow_static_const: bool,
     /// Whether a generated constant should be constexpr in C++ mode.
     pub allow_constexpr: bool,
+    /// Sort key for constants
+    pub sort_by: Option<SortKey>,
 }
 
 impl Default for ConstantConfig {
@@ -564,6 +702,7 @@ impl Default for ConstantConfig {
         ConstantConfig {
             allow_static_const: true,
             allow_constexpr: false,
+            sort_by: None,
         }
     }
 }
@@ -578,13 +717,34 @@ pub struct MacroExpansionConfig {
     pub bitflags: bool,
 }
 
-/// Settings to apply when running `rustc --pretty=expanded`
+/// Controls which Cargo profile is used for macro expansion.
+#[derive(Debug, Copy, Clone, PartialEq)]
+pub enum Profile {
+    Debug,
+    Release,
+}
+
+impl FromStr for Profile {
+    type Err = String;
+
+    fn from_str(s: &str) -> Result<Profile, Self::Err> {
+        match s {
+            "debug" | "Debug" => Ok(Profile::Debug),
+            "release" | "Release" => Ok(Profile::Release),
+            _ => Err(format!("Unrecognized Profile: '{}'.", s)),
+        }
+    }
+}
+
+deserialize_enum_str!(Profile);
+
+/// Settings to apply when running `rustc -Zunpretty=expanded`
 #[derive(Debug, Clone, Deserialize)]
 #[serde(rename_all = "snake_case")]
 #[serde(deny_unknown_fields)]
 #[serde(default)]
 pub struct ParseExpandConfig {
-    /// The names of crates to parse with `rustc --pretty=expanded`
+    /// The names of crates to parse with `rustc -Zunpretty=expanded`
     pub crates: Vec<String>,
     /// Whether to enable all the features when expanding.
     pub all_features: bool,
@@ -593,6 +753,8 @@ pub struct ParseExpandConfig {
     /// List of features to use when expanding. Combines with `default_features` like in
     /// `Cargo.toml`.
     pub features: Option<Vec<String>>,
+    /// Controls whether or not to pass `--release` when expanding.
+    pub profile: Profile,
 }
 
 impl Default for ParseExpandConfig {
@@ -602,6 +764,7 @@ impl Default for ParseExpandConfig {
             all_features: false,
             default_features: true,
             features: None,
+            profile: Profile::Debug,
         }
     }
 }
@@ -633,6 +796,7 @@ fn retrocomp_parse_expand_config_deserialize<'de, D: Deserializer<'de>>(
                 all_features: true,
                 default_features: true,
                 features: None,
+                profile: Profile::Debug,
             })
         }
 
@@ -663,10 +827,10 @@ pub struct ParseConfig {
     pub include: Option<Vec<String>>,
     /// The names of crates to not parse
     pub exclude: Vec<String>,
-    /// The configuration options for `rustc --pretty=expanded`
+    /// The configuration options for `rustc -Zunpretty=expanded`
     #[serde(deserialize_with = "retrocomp_parse_expand_config_deserialize")]
     pub expand: ParseExpandConfig,
-    /// Whether to use a new temporary target directory when running `rustc --pretty=expanded`.
+    /// Whether to use a new temporary target directory when running `rustc -Zunpretty=expanded`.
     /// This may be required for some build processes.
     pub clean: bool,
     /// List of crate names which generate consts, statics, and fns. By default
@@ -689,6 +853,29 @@ impl ParseConfig {
     }
 }
 
+/// Settings to apply to pointers
+#[derive(Debug, Clone, Default, Deserialize)]
+#[serde(rename_all = "snake_case")]
+#[serde(deny_unknown_fields)]
+#[serde(default)]
+pub struct PtrConfig {
+    /// Optional attribute to apply to pointers that are required to not be null
+    pub non_null_attribute: Option<String>,
+}
+
+/// Settings specific to Cython bindings.
+#[derive(Debug, Clone, Default, Deserialize)]
+#[serde(rename_all = "snake_case")]
+#[serde(deny_unknown_fields)]
+#[serde(default)]
+pub struct CythonConfig {
+    /// Header specified in the top level `cdef extern from header:` declaration.
+    pub header: Option<String>,
+    /// `from module cimport name1, name2, ...` declarations added in the same place
+    /// where you'd get includes in C.
+    pub cimports: BTreeMap<String, Vec<String>>,
+}
+
 /// A collection of settings to customize the generated bindings.
 #[derive(Debug, Clone, Deserialize)]
 #[serde(rename_all = "snake_case")]
@@ -701,10 +888,14 @@ pub struct Config {
     pub includes: Vec<String>,
     /// A list of additional system includes to put at the beginning of the generated header
     pub sys_includes: Vec<String>,
+    /// Optional verbatim code added after the include blocks
+    pub after_includes: Option<String>,
     /// Optional text to output at the end of the file
     pub trailer: Option<String>,
     /// Optional name to use for an include guard
     pub include_guard: Option<String>,
+    /// Add a `#pragma once` guard
+    pub pragma_once: bool,
     /// Generates no includes at all. Overrides all other include options
     ///
     /// This option is useful when using cbindgen with tools such as python's cffi which
@@ -726,12 +917,19 @@ pub struct Config {
     pub line_length: usize,
     /// The amount of spaces in a tab
     pub tab_width: usize,
+    /// The type of line endings to generate
+    pub line_endings: LineEndingStyle,
     /// The language to output bindings for
     pub language: Language,
     /// Include preprocessor defines in C bindings to ensure C++ compatibility
     pub cpp_compat: bool,
     /// The style to declare structs, enums and unions in for C
     pub style: Style,
+    /// Default sort key for functions and constants.
+    pub sort_by: SortKey,
+    /// If this option is true `usize` and `isize` will be converted into `size_t` and `ptrdiff_t`
+    /// instead of `uintptr_t` and `intptr_t` respectively.
+    pub usize_is_size_t: bool,
     /// The configuration options for parsing
     pub parse: ParseConfig,
     /// The configuration options for exporting
@@ -754,10 +952,52 @@ pub struct Config {
     pub constant: ConstantConfig,
     /// Preprocessor defines to use when generating #ifdef's for #[cfg]
     pub defines: HashMap<String, String>,
-    /// Include doc comments from rust as documentation
+    /// Include doc comments from Rust as documentation
     pub documentation: bool,
     /// How documentation comments should be styled.
     pub documentation_style: DocumentationStyle,
+    /// How much of the documentation should be output for each item.
+    pub documentation_length: DocumentationLength,
+    /// Configuration options for pointers
+    #[serde(rename = "ptr")]
+    pub pointer: PtrConfig,
+    /// Only download sources for dependencies needed for the target platform.
+    ///
+    /// By default, cbindgen will fetch sources for dependencies used on any platform so that if a
+    /// type is defined in terms of a type from a dependency on another target (probably behind a
+    /// `#[cfg]`), cbindgen will be able to generate the appropriate binding as it can see the
+    /// nested type's definition. However, this makes calling cbindgen slower, as it may have to
+    /// download a number of additional dependencies.
+    ///
+    /// As an example, consider this Cargo.toml:
+    ///
+    /// ```toml
+    /// [target.'cfg(windows)'.dependencies]
+    /// windows = "0.7"
+    /// ```
+    ///
+    /// with this declaration in one of the `.rs` files that cbindgen is asked to generate bindings
+    /// for:
+    ///
+    /// ```rust,ignore
+    /// #[cfg(windows)]
+    /// pub struct Error(windows::ErrorCode);
+    /// ```
+    ///
+    /// With the default value (`false`), cbindgen will download the `windows` dependency even when
+    /// not compiling for Windows, and will thus be able to generate the binding for `Error`
+    /// (behind a `#define`).
+    ///
+    /// If this value is instead to `true`, cbindgen will _not_ download the `windows` dependency
+    /// if it's not compiling for Windows, but will also fail to generate a Windows binding for
+    /// `Error` as it does not know the definition for `ErrorCode`.
+    ///
+    /// The target can be chosen via the `TARGET` environment variable (if used
+    /// via the CLI, when ran from a build script cargo sets this variable
+    /// appropriately).
+    pub only_target_dependencies: bool,
+    /// Configuration options specific to Cython.
+    pub cython: CythonConfig,
 }
 
 impl Default for Config {
@@ -766,8 +1006,10 @@ impl Default for Config {
             header: None,
             includes: Vec::new(),
             sys_includes: Vec::new(),
+            after_includes: None,
             trailer: None,
             include_guard: None,
+            pragma_once: false,
             autogen_warning: None,
             include_version: false,
             no_includes: false,
@@ -777,9 +1019,12 @@ impl Default for Config {
             braces: Braces::SameLine,
             line_length: 100,
             tab_width: 2,
+            line_endings: LineEndingStyle::default(),
             language: Language::Cxx,
             cpp_compat: false,
-            style: Style::Type,
+            style: Style::default(),
+            usize_is_size_t: false,
+            sort_by: SortKey::None,
             macro_expansion: Default::default(),
             parse: ParseConfig::default(),
             export: ExportConfig::default(),
@@ -791,17 +1036,49 @@ impl Default for Config {
             defines: HashMap::new(),
             documentation: true,
             documentation_style: DocumentationStyle::Auto,
+            documentation_length: DocumentationLength::Full,
+            pointer: PtrConfig::default(),
+            only_target_dependencies: false,
+            cython: CythonConfig::default(),
         }
     }
 }
 
 impl Config {
+    pub(crate) fn cpp_compatible_c(&self) -> bool {
+        self.language == Language::C && self.cpp_compat
+    }
+
+    pub(crate) fn include_guard(&self) -> Option<&str> {
+        if self.language == Language::Cython {
+            None
+        } else {
+            self.include_guard.as_deref()
+        }
+    }
+
+    pub(crate) fn includes(&self) -> &[String] {
+        if self.language == Language::Cython {
+            &[]
+        } else {
+            &self.includes
+        }
+    }
+
+    pub(crate) fn sys_includes(&self) -> &[String] {
+        if self.language == Language::Cython {
+            &[]
+        } else {
+            &self.sys_includes
+        }
+    }
+
     pub fn from_file<P: AsRef<StdPath>>(file_name: P) -> Result<Config, String> {
-        let config_text = fs::read_to_string(file_name.as_ref()).or_else(|_| {
-            Err(format!(
+        let config_text = fs::read_to_string(file_name.as_ref()).map_err(|_| {
+            format!(
                 "Couldn't open config file: {}.",
                 file_name.as_ref().display()
-            ))
+            )
         })?;
 
         match toml::from_str::<Config>(&config_text) {
