@@ -6,6 +6,7 @@ use std::borrow::Cow;
 use std::collections::HashMap;
 use std::io::Write;
 
+use syn::ext::IdentExt;
 use syn::{self, UnOp};
 
 use crate::bindgen::config::{Config, Language};
@@ -65,7 +66,7 @@ impl Literal {
                 if path.replace_self_with(self_ty) {
                     *export_name = self_ty.name().to_owned();
                 }
-                for (ref _name, ref mut expr) in fields {
+                for ref mut expr in fields.values_mut() {
                     expr.replace_self_with(self_ty);
                 }
             }
@@ -122,7 +123,7 @@ impl Literal {
                 ..
             } => {
                 config.export.rename(export_name);
-                for (_, lit) in fields {
+                for lit in fields.values_mut() {
                     lit.rename_for_config(config);
                 }
             }
@@ -195,16 +196,13 @@ impl Literal {
                 })
             }
 
-            // Match literals like "one", 'a', 32 etc
+            // Match literals like true, 'a', 32 etc
             syn::Expr::Lit(syn::ExprLit { ref lit, .. }) => {
                 match lit {
-                    syn::Lit::Str(ref value) => {
-                        Ok(Literal::Expr(format!("u8\"{}\"", value.value())))
-                    }
                     syn::Lit::Byte(ref value) => Ok(Literal::Expr(format!("{}", value.value()))),
                     syn::Lit::Char(ref value) => Ok(Literal::Expr(match value.value() as u32 {
                         0..=255 => format!("'{}'", value.value().escape_default()),
-                        other_code => format!(r"L'\U{:08X}'", other_code),
+                        other_code => format!(r"U'\U{:08X}'", other_code),
                     })),
                     syn::Lit::Int(ref value) => {
                         if value.base10_parse::<i64>().is_err() {
@@ -227,11 +225,11 @@ impl Literal {
                 ref fields,
                 ..
             }) => {
-                let struct_name = path.segments[0].ident.to_string();
+                let struct_name = path.segments[0].ident.unraw().to_string();
                 let mut field_map = HashMap::<String, Literal>::default();
                 for field in fields {
                     let ident = match field.member {
-                        syn::Member::Named(ref name) => name.to_string(),
+                        syn::Member::Named(ref name) => name.unraw().to_string(),
                         syn::Member::Unnamed(ref index) => format!("_{}", index.index),
                     };
                     let key = ident.to_string();
@@ -290,9 +288,13 @@ impl Literal {
         }
     }
 
-    fn write<F: Write>(&self, config: &Config, out: &mut SourceWriter<F>) {
+    pub(crate) fn write<F: Write>(&self, config: &Config, out: &mut SourceWriter<F>) {
         match self {
-            Literal::Expr(v) => write!(out, "{}", v),
+            Literal::Expr(v) => match (&**v, config.language) {
+                ("true", Language::Cython) => write!(out, "True"),
+                ("false", Language::Cython) => write!(out, "False"),
+                (v, _) => write!(out, "{}", v),
+            },
             Literal::Path(v) => write!(out, "{}", v),
             Literal::PostfixUnaryOp { op, ref value } => {
                 write!(out, "{}", op);
@@ -310,9 +312,17 @@ impl Literal {
                 write!(out, ")");
             }
             Literal::Cast { ref ty, ref value } => {
-                write!(out, "(");
+                out.write(if config.language == Language::Cython {
+                    "<"
+                } else {
+                    "("
+                });
                 ty.write(config, out);
-                write!(out, ")");
+                out.write(if config.language == Language::Cython {
+                    ">"
+                } else {
+                    ")"
+                });
                 value.write(config, out);
             }
             Literal::Struct {
@@ -320,10 +330,10 @@ impl Literal {
                 fields,
                 path,
             } => {
-                if config.language == Language::C {
-                    write!(out, "({})", export_name);
-                } else {
-                    write!(out, "{}", export_name);
+                match config.language {
+                    Language::C => write!(out, "({})", export_name),
+                    Language::Cxx => write!(out, "{}", export_name),
+                    Language::Cython => write!(out, "<{}>", export_name),
                 }
 
                 write!(out, "{{ ");
@@ -331,16 +341,16 @@ impl Literal {
                 // In C++, same order as defined is required.
                 let ordered_fields = out.bindings().struct_field_names(path);
                 for ordered_key in ordered_fields.iter() {
-                    if let Some(ref lit) = fields.get(ordered_key) {
+                    if let Some(lit) = fields.get(ordered_key) {
                         if !is_first_field {
                             write!(out, ", ");
                         } else {
                             is_first_field = false;
                         }
-                        if config.language == Language::Cxx {
-                            write!(out, "/* .{} = */ ", ordered_key);
-                        } else {
-                            write!(out, ".{} = ", ordered_key);
+                        match config.language {
+                            Language::Cxx => write!(out, "/* .{} = */ ", ordered_key),
+                            Language::C => write!(out, ".{} = ", ordered_key),
+                            Language::Cython => {}
                         }
                         lit.write(config, out);
                     }
@@ -363,16 +373,6 @@ pub struct Constant {
     pub associated_to: Option<Path>,
 }
 
-fn can_handle(ty: &Type, expr: &syn::Expr) -> bool {
-    if ty.is_primitive_or_ptr_primitive() {
-        return true;
-    }
-    match *expr {
-        syn::Expr::Struct(_) => true,
-        _ => false,
-    }
-}
-
 impl Constant {
     pub fn load(
         path: Path,
@@ -390,11 +390,7 @@ impl Constant {
             }
         };
 
-        if !can_handle(&ty, expr) {
-            return Err("Unhandled const definition".to_owned());
-        }
-
-        let mut lit = Literal::load(&expr)?;
+        let mut lit = Literal::load(expr)?;
 
         if let Some(ref associated_to) = associated_to {
             ty.replace_self_with(associated_to);
@@ -494,7 +490,7 @@ impl Constant {
         debug_assert!(config.structure.associated_constants_in_body);
         debug_assert!(config.constant.allow_static_const);
 
-        if let Type::ConstPtr(..) = self.ty {
+        if let Type::Ptr { is_const: true, .. } = self.ty {
             out.write("static ");
         } else {
             out.write("static const ");
@@ -556,7 +552,7 @@ impl Constant {
                 ref fields,
                 ref path,
                 ..
-            } if out.bindings().struct_is_transparent(path) => &fields.iter().next().unwrap().1,
+            } if out.bindings().struct_is_transparent(path) => fields.iter().next().unwrap().1,
             _ => &self.value,
         };
 
@@ -568,31 +564,41 @@ impl Constant {
             false
         };
 
-        if (config.constant.allow_static_const || allow_constexpr)
-            && config.language == Language::Cxx
-        {
-            if allow_constexpr {
-                out.write("constexpr ")
-            }
+        match config.language {
+            Language::Cxx if config.constant.allow_static_const || allow_constexpr => {
+                if allow_constexpr {
+                    out.write("constexpr ")
+                }
 
-            if config.constant.allow_static_const {
-                out.write(if in_body { "inline " } else { "static " });
-            }
+                if config.constant.allow_static_const {
+                    out.write(if in_body { "inline " } else { "static " });
+                }
 
-            if let Type::ConstPtr(..) = self.ty {
-                // Nothing.
-            } else {
+                if let Type::Ptr { is_const: true, .. } = self.ty {
+                    // Nothing.
+                } else {
+                    out.write("const ");
+                }
+
+                self.ty.write(config, out);
+                write!(out, " {} = ", name);
+                value.write(config, out);
+                write!(out, ";");
+            }
+            Language::Cxx | Language::C => {
+                write!(out, "#define {} ", name);
+                value.write(config, out);
+            }
+            Language::Cython => {
                 out.write("const ");
+                self.ty.write(config, out);
+                // For extern Cython declarations the initializer is ignored,
+                // but still useful as documentation, so we write it as a comment.
+                write!(out, " {} # = ", name);
+                value.write(config, out);
             }
-
-            self.ty.write(config, out);
-            write!(out, " {} = ", name);
-            value.write(config, out);
-            write!(out, ";");
-        } else {
-            write!(out, "#define {} ", name);
-            value.write(config, out);
         }
+
         condition.write_after(config, out);
     }
 }
